#include <sps2lib.h>
#include <sps2tags.h>
#include <sps2util.h>
#include <sps2regstructs.h>
#include <assert.h>
#include <memory.h>
#include <signal.h>
#include "PS2Defines.h"
#include "ps2maths.h"
#include "ps2matrix4x4.h"
#include "pad.h"
#include "sps2wrap.h"
#include "dma.h"
#include "texture.h"
#include "PS2Types.h"

//function prototypes in this file
void MoveSprite(PS2Sprite_t * pSP, float Delta, int Axis);
void AnimateSprite(PS2Sprite_t * pSP, float Count);
void BuildSprite2D(PS2Sprite_t * pSprite);
void GameLogic(PS2Sprite_t * pSP);

bool g_bLoop = true;

void sig_handle(int sig)
{
	g_bLoop = false;
}

int main()
{
	PS2Sprite_t Sprite[1];		//some sprite structures to play with
	
	// Initialise control pad 0
	if(!pad_init(PAD_0, PAD_INIT_LOCK | PAD_INIT_ANALOGUE | PAD_INIT_PRESSURE))
	{
		printf("Failed to initialise control pad\n");
		pad_cleanup(PAD_0);
		exit(0);
	}	

	SPS2Manager.Initialise(4096);	// 4096 * 4K Pages = 16MB Total

	VIFStaticDMA.Initialise(1536);	// 1536 * 4K Pages = 6MB Static DMA
	VIFDynamicDMA.Initialise(256);	// 256 * 4K Pages * 2 Buffers =
									// 2MB Dynamic DMA

	// Register our signal function for every possible signal (i.e. ctrl + c)
	for(int sig=0; sig<128; sig++)
		signal(sig, sig_handle);
		
	// Set up the DMA packet to clear the screen.
	SPS2Manager.InitScreenClear(0x10, 0x10, 0x20);	
		
	// Load texture.
	CTexture Tex1;
	Tex1.LoadBitmap("game.bmp", true, false);
	
	// Configure the Sprite Structures
	
	// Set world matrix to identity
	Sprite[0].World = Matrix4x4::IDENTITY;
	
	Sprite[0].Pos.x = 0.0f;
	Sprite[0].Pos.y = 0.0f;
	Sprite[0].Pos.z = 100.0f;
	Sprite[0].Pos.w = 1.0f;
	Sprite[0].w = 32;
	Sprite[0].h = 32;
	
	
	// Untransformed Sprite positions
	// Top Left
	Sprite[0].TL.x = -(float)(Sprite[0].w/2);;
	Sprite[0].TL.y = -(float)(Sprite[0].h/2);
	Sprite[0].TL.z = 0.0f;
	Sprite[0].TL.w = 1.0f;
	// Bottom Left
	Sprite[0].BL.x = -(float)(Sprite[0].w/2);
	Sprite[0].BL.y =  (float)(Sprite[0].h/2);
	Sprite[0].BL.z = 0.0f;
	Sprite[0].BL.w = 1.0f;
	// Top Right
	Sprite[0].TR.x =  (float)(Sprite[0].w/2);
	Sprite[0].TR.y = -(float)(Sprite[0].h/2);
	Sprite[0].TR.z = 0.0f;
	Sprite[0].TR.w = 1.0f;
	// Bottom Right
	Sprite[0].BR.x = (float)(Sprite[0].w/2);
	Sprite[0].BR.y = (float)(Sprite[0].h/2);
	Sprite[0].BR.z = 0.0f;
	Sprite[0].BR.w = 1.0f;
	// UV Texture offsets
	Sprite[0].u = 32;
	Sprite[0].v = 0;
	// Colour
	Sprite[0].r = 0x80;
	Sprite[0].g = 0x80;
	Sprite[0].b = 0x80;
	Sprite[0].a = 0x80;
	// Animation parameters
	Sprite[0].FCountMax = 5;	
	Sprite[0].FCount = 0;
	Sprite[0].NFrames = 8;
	// Rotation angle
	Sprite[0].RotX = 0.0f;
	Sprite[0].RotY = 0.0f;
	Sprite[0].RotZ = 0.0f;	

	while(g_bLoop)
	{
		// Fire off the previous frame's buffer
		VIFDynamicDMA.Fire();

		pad_update(PAD_0);
		
		// Perform the logic for the game
		GameLogic(Sprite);
	
		// Clear the screen
		SPS2Manager.BeginScene();
		
		// Upload the current texture to the free texture buffer
		Tex1.Upload(480);
		Tex1.Select();
		
		// Start DIRECT mode transfer
		VIFDynamicDMA.StartDirect();
		
		//build the GS packet for all objects to be drawn
		BuildSprite2D(&Sprite[0]);
		
		// End the direct mode
		VIFDynamicDMA.EndDirect();
	
		// Swap the buffers
		SPS2Manager.EndScene();
		
		// Check for exit condition
		if((pad[0].buttons & PAD_START)&&(pad[0].buttons & PAD_SELECT)) g_bLoop = false;
	}
	
	// Cleanup the Pads
	pad_cleanup(PAD_0);
	
	return 0;
}




void BuildSprite2D(PS2Sprite_t * pSprite)
{	
	
	// Get the rotation matrix from the rotation angle
	Matrix4x4 RotX, RotY, RotZ;
	
	RotX.RotationX(pSprite->RotX);
	RotY.RotationY(pSprite->RotY);
	RotZ.RotationZ(pSprite->RotZ);
	
	pSprite->World = RotZ * RotY * RotX;
		
	// These are the transformed coordinates of the sprite
	Vector4 TR, TL, BR, BL;

	//Add the position to the world matrix
	pSprite->World(3,0) = pSprite->Pos.x;
	pSprite->World(3,1) = pSprite->Pos.y;
	pSprite->World(3,2) = pSprite->Pos.z;	//this is the z depth
		
	// Apply World matrix to the untransformed coordinates
	TR = pSprite->World * pSprite->TR;
	TL = pSprite->World * pSprite->TL;
	BR = pSprite->World * pSprite->BR;
	BL = pSprite->World * pSprite->BL;
	
			
	// Create a textured tristrip
	// Add the GIFTag
	VIFDynamicDMA.Add128(GS_GIFTAG_BATCH(	4,							// NLOOP
	 										1,							// EOP
	 										1,							// PRE
	 										GS_PRIM(PRIM_TRI_STRIP, 	// PRIM
													PRIM_IIP_FLAT, 
													PRIM_TME_ON,
													PRIM_FGE_OFF, 
													PRIM_ABE_ON, 
													PRIM_AA1_OFF, 
													PRIM_FST_UV, 
													PRIM_CTXT_CONTEXT1, 
													PRIM_FIX_NOFIXDDA),					
	 										GIF_FLG_PACKED,				// FLG
	 										GS_BATCH_3(	GIF_REG_RGBAQ, 	// BATCH
	 													GIF_REG_UV, 
	 													GIF_REG_XYZ2)));

		// Add Vertex data
		
		//top left
		// Add the colour
    	VIFDynamicDMA.Add128(PACKED_RGBA(pSprite->r, pSprite->g, pSprite->b, pSprite->a));
    	
    	// Add UV texture coordinates
    	VIFDynamicDMA.Add128(PACKED_UV(	((pSprite->u)<<4), 
    									((pSprite->v)<<4)));
    	    	
    	// Add the position  
    	// The (1<<3) is a half-pixel offset to get the texture and sprite to line-up 	   	
    	VIFDynamicDMA.Add128(PACKED_XYZ2(	(2048 << 4) + (int)((TL.x)*16.0f) + (1<<3), 
    										(2048 << 4) + (int)((TL.y)*16.0f) + (1<<3), 
    										(int)TL.z, 
    										0)); 													

		//bottom left
		// Add the colour
    	VIFDynamicDMA.Add128(PACKED_RGBA(pSprite->r, pSprite->g, pSprite->b, pSprite->a));
    	
    	// Add UV texture coordinates
    	VIFDynamicDMA.Add128(PACKED_UV(	((pSprite->u)<<4), 
    									(((pSprite->v + pSprite->h)<<4))));
    	    	
    	// Add the position  
    	// The (1<<3) is a half-pixel offset to get the texture and sprite to line-up 	   	
    	VIFDynamicDMA.Add128(PACKED_XYZ2(	(2048 << 4) + (int)((BL.x)*16.0f) + (1<<3), 
    										(2048 << 4) + (int)((BL.y)*16.0f) + (1<<3), 
    										(int)BL.z, 
    										0)); 		
    	
    	//top right
    	// Add the colour
    	VIFDynamicDMA.Add128(PACKED_RGBA(pSprite->r, pSprite->g, pSprite->b, pSprite->a));
    	
    	// Add UV texture coordinates
    	VIFDynamicDMA.Add128(PACKED_UV(	((pSprite->u + pSprite->w)<<4), 
    									((pSprite->v)<<4)));
    	    	
    	// Add the position  
    	// The (1<<3) is a half-pixel offset to get the texture and sprite to line-up 	   	
    	VIFDynamicDMA.Add128(PACKED_XYZ2(	(2048 << 4) + (int)((TR.x)*16.0f) + (1<<3), 
    										(2048 << 4) + (int)((TR.y)*16.0f) + (1<<3), 
    										(int)TR.z, 
    										0)); 		
    	
    	// Bottom Right
    	// Add the colour
    	VIFDynamicDMA.Add128(PACKED_RGBA(pSprite->r, pSprite->g, pSprite->b, pSprite->a));
    	
    	// Add UV texture coordinates
    	VIFDynamicDMA.Add128(PACKED_UV(	((pSprite->u + pSprite->w)<<4), 
    									((pSprite->v + pSprite->h)<<4)));
    	    	
    	// Add the position  
    	// The (1<<3) is a half-pixel offset to get the texture and sprite to line-up 	   	
    	VIFDynamicDMA.Add128(PACKED_XYZ2(	(2048 << 4) + (int)((BR.x)*16.0f) + (1<<3), 
    										(2048 << 4) + (int)((BR.y)*16.0f) + (1<<3), 
    										(int)BR.z, 
    										0)); 		
    										
   										
}


//move sprite in the direction it is pointing
void MoveSprite(PS2Sprite_t * pSP, float Delta, int Axis)
{
	//this moves the sprite along its local "Axis" axis
	// by an amount Delta
	pSP->Pos.x += pSP->World(Axis,0)* Delta;
	pSP->Pos.y += pSP->World(Axis,1)* Delta;
	pSP->Pos.z += pSP->World(Axis,2)* Delta;
}


//some very simple animation
void AnimateSprite(PS2Sprite_t * pSP, float Count)
{
	int FCountMax = 8 - (int)(6.0f * Count);
	if((pSP->FCount++)%FCountMax == 0)
		{
			pSP->v += pSP->h;
			if(pSP->v > (pSP->NFrames - 1) * pSP->h)pSP->v = 0;
		}
}


// Perform the logic for the game
void GameLogic(PS2Sprite_t * pSP)
{
	//rotate the sprite round z axis
	if(pad[0].buttons & PAD_RIGHT)pSP[0].RotZ += pad[0].pressures[PAD_PRIGHT] * 0.2f;
	if(pad[0].buttons & PAD_LEFT) pSP[0].RotZ -= pad[0].pressures[PAD_PLEFT] * 0.2f;
	
	//rotate the sprite round y axis
	if(pad[0].buttons & PAD_UP)pSP[0].RotY += pad[0].pressures[PAD_PUP] * 0.2f;
	if(pad[0].buttons & PAD_DOWN) pSP[0].RotY -= pad[0].pressures[PAD_PDOWN] * 0.2f;
	
	//rotate the sprite round x axis
	if(pad[0].buttons & PAD_L1)pSP[0].RotX += pad[0].pressures[PAD_PL1] * 0.2f;
	if(pad[0].buttons & PAD_L2) pSP[0].RotX -= pad[0].pressures[PAD_PL2] * 0.2f;

	//move along local x and animate the sprite
	if(pad[0].buttons & PAD_SQUARE)
	{
		MoveSprite(&pSP[0], -pad[0].pressures[PAD_PSQUARE] * 3.0f, XAXIS);
		AnimateSprite(&pSP[0], pad[0].pressures[PAD_PSQUARE]);
	}
		
	if(pad[0].buttons & PAD_CIRCLE)
	{
		MoveSprite(&pSP[0], pad[0].pressures[PAD_PCIRCLE] * 3.0f, XAXIS);
		AnimateSprite(&pSP[0], pad[0].pressures[PAD_PCIRCLE]);
	}
		
	//move along local y and animate the sprite
	if(pad[0].buttons & PAD_TRI)
	{
		MoveSprite(&pSP[0], pad[0].pressures[PAD_PTRI] * 3.0f, YAXIS);
		AnimateSprite(&pSP[0], pad[0].pressures[PAD_PTRI]);
	}
		
	if(pad[0].buttons & PAD_CROSS)
	{
		MoveSprite(&pSP[0], -pad[0].pressures[PAD_PCROSS] * 3.0f, YAXIS);
		AnimateSprite(&pSP[0], pad[0].pressures[PAD_PCROSS]);
	}
	
	//move along local z and animate the sprite
	if(pad[0].buttons & PAD_R1)
	{
		MoveSprite(&pSP[0], pad[0].pressures[PAD_PR1] * 3.0f, ZAXIS);
		AnimateSprite(&pSP[0], pad[0].pressures[PAD_PR1]);
	}
	
	if(pad[0].buttons & PAD_R2)
	{
		MoveSprite(&pSP[0], -pad[0].pressures[PAD_PR2] * 3.0f, ZAXIS);
		AnimateSprite(&pSP[0], pad[0].pressures[PAD_PR2]);
	}
}